import asyncio
import time
from collections.abc import Sequence
from typing import Annotated as A
from typing import TypeVar, get_origin

import numpy as np
import pytest
from tango.asyncio import DeviceProxy
from test_base_device import TestDevice

from ophyd_async.core import NotConnectedError, SignalRW, StandardReadable, StrictEnum
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.tango.core import (
    DevStateEnum,
    TangoDevice,
    TangoSignalBackend,
    get_full_attr_trl,
    infer_python_type,
    parse_precision,
    tango_signal_r,
    tango_signal_rw,
    tango_signal_w,
    tango_signal_x,
)
from ophyd_async.tango.testing import (
    ExampleStrEnum,
    OneOfEverythingTangoDevice,
)
from ophyd_async.testing import (
    MonitorQueue,
    assert_reading,
    assert_value,
    partial_reading,
)

T = TypeVar("T")


# --------------------------------------------------------------------
#               TestDevice
# --------------------------------------------------------------------
# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def tango_test_device(subprocess_helper):
    with subprocess_helper(
        [{"class": TestDevice, "devices": [{"name": "test/device/1"}]}]
    ) as context:
        yield context.trls["test/device/1"]


# --------------------------------------------------------------------
def assert_enum(initial_value, readout_value):
    if type(readout_value) in [list, tuple]:
        for _initial_value, _readout_value in zip(
            initial_value, readout_value, strict=False
        ):
            assert_enum(_initial_value, _readout_value)
    else:
        assert initial_value == readout_value


# --------------------------------------------------------------------
#               fixtures to run Echo device
# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def everything_device_trl(subprocess_helper):
    with subprocess_helper(
        [{"class": OneOfEverythingTangoDevice, "devices": [{"name": "test/device/2"}]}]
    ) as context:
        yield context.trls["test/device/2"]


class TangoEverythingOphydDevice(TangoDevice, StandardReadable):
    # datatype of enum commands must be explicitly hinted
    strenum_cmd: A[SignalRW[ExampleStrEnum], Format.HINTED_UNCACHED_SIGNAL]


@pytest.fixture()
async def everything_device(everything_device_trl):
    return TangoEverythingOphydDevice(everything_device_trl)


# --------------------------------------------------------------------
#               helpers to run tests
# --------------------------------------------------------------------
def get_test_descriptor(python_type: type[T], value: T, is_cmd: bool) -> dict:
    if python_type in [bool]:
        return {"dtype": "boolean", "shape": []}
    if python_type in [int]:
        return {"dtype": "integer", "shape": []}
    if python_type in [float]:
        return {"dtype": "number", "shape": []}
    if python_type in [str]:
        return {"dtype": "string", "shape": []}
    if get_origin(python_type) is Sequence:
        return {"dtype": "array", "shape": [len(value)]}
    if issubclass(python_type, StrictEnum):
        return {"dtype": "string", "shape": []}
    return {
        "dtype": "array",
        "shape": [np.iinfo(np.int32).max] if is_cmd else list(np.array(value).shape),
    }


# --------------------------------------------------------------------
async def make_backend(
    typ: type | None,
    pv: str,
    connect: bool = True,
    allow_events: bool | None = True,
) -> TangoSignalBackend:
    """Wrapper for making the tango signal backend."""

    backend = TangoSignalBackend(typ, pv, pv)
    backend.allow_events(allow_events)
    if connect:
        await backend.connect(1)
    return backend


# --------------------------------------------------------------------
async def assert_monitor_then_put(
    signal: SignalRW,
    source: str,
    initial_value: T,
    put_value: T,
    descriptor: dict,
):
    backend = signal._connector.backend
    # Make a monitor queue that will monitor for updates
    with MonitorQueue(signal) as q:
        test_descriptor = dict(source=source, **descriptor)
        try:
            backend_datakey = await backend.get_datakey(source)
        except Exception as exc:
            pytest.fail(f"Failed to get datakey for {source}: {exc}")
        for key, value in test_descriptor.items():
            assert backend_datakey[key] == value, (
                f"Key {key} mismatch: {value} != {backend_datakey[key]}."
                f" Source: {source}"
            )
        # Check initial value
        await q.assert_updates(initial_value)
        # Put to new value and check that
        await backend.put(put_value)
        await q.assert_updates(put_value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(18.8)
async def test_backend_get_put_monitor_attr(
    everything_device: TangoDevice, everything_signal_info
):
    await everything_device.connect()
    try:
        for attr_data in everything_signal_info.values():
            signal = getattr(everything_device, attr_data.name)
            source = get_full_attr_trl(everything_device._connector.trl, attr_data.name)
            await asyncio.wait_for(
                assert_monitor_then_put(
                    signal,
                    source,
                    attr_data.initial,
                    attr_data.random_value(),
                    get_test_descriptor(attr_data.py_type, attr_data.initial, False),
                ),
                timeout=10,  # Timeout in seconds
            )
    except TimeoutError:
        pytest.fail("Test timed out")
    except Exception as exc:
        pytest.fail(f"Test failed with exception: {exc}")


# --------------------------------------------------------------------
async def assert_put_read(
    signal: SignalRW,
    source: str,
    put_value: T,
    datatype: type[T] | None = None,  # TODO reimplement this
):
    backend = signal._connector.backend
    await backend.put(put_value)

    expected_reading = {
        "timestamp": pytest.approx(time.time(), rel=0.1),
        "alarm_severity": 0,
    }

    get_reading = dict(await backend.get_reading())
    get_reading.pop("value")
    assert expected_reading == get_reading


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_backend_get_put_monitor_cmd(
    everything_device: TangoDevice, everything_signal_info
):
    await everything_device.connect()
    for cmd_data in everything_signal_info.values():
        if cmd_data.cmd_name is None:
            continue
        put_value = cmd_data.random_value()
        # With the given datatype, check we have the correct initial value
        # and putting works
        signal = getattr(everything_device, cmd_data.cmd_name)
        source = get_full_attr_trl(everything_device._connector.trl, cmd_data.cmd_name)
        await assert_put_read(signal, source, put_value, cmd_data.py_type)
        # # With guessed datatype, check we can set it back to the initial value
        await assert_put_read(signal, source, put_value)
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        await asyncio.gather(*tasks)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_signal_r(everything_device_trl: str, everything_signal_info):
    timeout = 0.2
    for attr_data in everything_signal_info.values():
        source = get_full_attr_trl(everything_device_trl, attr_data.name)
        signal = tango_signal_r(
            datatype=attr_data.py_type,
            read_trl=source,
            timeout=timeout,
            name="test_signal",
        )
        await signal.connect()
        # value may have changed from initial value if other tests run first
        # casting to array makes tuples returned from tango for arrays work with
        # assert_* functions
        value = np.array(await signal.get_value())
        await assert_value(signal, value)
        await assert_reading(signal, {"test_signal": partial_reading(value)})


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_signal_w(everything_device_trl: str, everything_signal_info):
    timeout = 0.2
    for attr_data in everything_signal_info.values():
        source = get_full_attr_trl(everything_device_trl, attr_data.name)
        signal = tango_signal_w(
            datatype=attr_data.py_type,
            write_trl=source,
            timeout=timeout,
            name="test_signal",
        )
        await signal.connect()  # have to connect to get correct converter

        put_value = attr_data.random_value()

        status = signal.set(put_value, timeout=timeout)
        await status
        assert status.done is True and status.success is True

        status = signal.set(put_value, timeout=timeout)
        await status
        assert status.done is True and status.success is True

        status = signal.set(put_value)
        await status
        assert status.done is True and status.success is True

        status = signal.set(put_value)
        await status
        assert status.done is True and status.success is True


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_signal_rw(everything_device_trl: str, everything_signal_info):
    timeout = 0.2
    for attr_data in everything_signal_info.values():
        source = get_full_attr_trl(everything_device_trl, attr_data.name)
        signal = tango_signal_rw(
            datatype=attr_data.py_type,
            read_trl=source,
            write_trl=source,
            timeout=timeout,
            name="test_signal",
        )
        await signal.connect()
        put_value = attr_data.random_value()
        await signal.set(put_value)
        await assert_value(signal, put_value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.timeout(2.0)
async def test_tango_signal_x(tango_test_device: str):
    timeout = 0.2
    signal = tango_signal_x(
        write_trl=get_full_attr_trl(tango_test_device, "clear"),
        timeout=timeout,
        name="test_signal",
    )
    await signal.connect()
    status = signal.trigger()
    await status
    assert status.done is True and status.success is True


async def assert_val_reading(signal, value, name=""):
    await assert_value(signal, value)
    await assert_reading(signal, {name: partial_reading(value)})


async def test_set_with_converter(everything_device_trl):
    everything_device = TangoDevice(everything_device_trl)
    await everything_device.connect()
    with pytest.raises(TypeError):
        await everything_device.strenum.set(0)
    with pytest.raises(ValueError):
        await everything_device.strenum.set("NON_ENUM_VALUE")
    await everything_device.strenum.set("AAA")
    await everything_device.strenum.set(ExampleStrEnum.B)
    await everything_device.strenum.set(ExampleStrEnum.C.value)
    # setting enum spectrum works with lists and arrays
    await everything_device.strenum_spectrum.set(["AAA", "BBB"])
    await everything_device.strenum_spectrum.set(np.array(["BBB", "CCC"]))
    await everything_device.strenum_spectrum.set(
        [
            ExampleStrEnum.B.value,
            ExampleStrEnum.C.value,
        ]
    )
    await everything_device.strenum_spectrum.set(
        np.array(
            [
                ExampleStrEnum.A,
                ExampleStrEnum.B,
            ],
            dtype=ExampleStrEnum,
            # when using enum instances, must use array with correct dtype
            # passing this as a list will cast the strings incorrectly
        )
    )

    await everything_device.my_state.set(DevStateEnum.EXTRACT)
    await everything_device.my_state_spectrum.set(
        np.array(
            [
                DevStateEnum.OPEN,
                DevStateEnum.CLOSE,
                DevStateEnum.MOVING,
            ],
            dtype=DevStateEnum,
        )
    )


@pytest.mark.timeout(18.8)
async def test_assert_val_reading_everything_tango(
    everything_device_trl, everything_signal_info
):
    esi = everything_signal_info
    everything_device = TangoDevice(everything_device_trl)
    await everything_device.connect()
    await everything_device.reset_values.trigger()
    await asyncio.sleep(1)
    await assert_val_reading(everything_device.str, esi["str"].initial)
    await assert_val_reading(everything_device.bool, esi["bool"].initial)
    await assert_val_reading(everything_device.strenum, esi["strenum"].initial)
    await assert_val_reading(everything_device.int8, esi["int8"].initial)
    await assert_val_reading(everything_device.uint8, esi["uint8"].initial)
    await assert_val_reading(everything_device.int16, esi["int16"].initial)
    await assert_val_reading(everything_device.uint16, esi["uint16"].initial)
    await assert_val_reading(everything_device.int32, esi["int32"].initial)
    await assert_val_reading(everything_device.uint32, esi["uint32"].initial)
    await assert_val_reading(everything_device.int64, esi["int64"].initial)
    await assert_val_reading(everything_device.uint64, esi["uint64"].initial)
    await assert_val_reading(everything_device.float32, esi["float32"].initial)
    await assert_val_reading(everything_device.float64, esi["float64"].initial)
    await assert_val_reading(everything_device.my_state, esi["my_state"].initial)

    await assert_val_reading(
        everything_device.str_spectrum, esi["str_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.bool_spectrum, esi["bool_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.strenum_spectrum, esi["strenum_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.int8_spectrum, esi["int8_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.uint8_spectrum, esi["uint8_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.int16_spectrum, esi["int16_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.uint16_spectrum, esi["uint16_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.int32_spectrum, esi["int32_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.uint32_spectrum, esi["uint32_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.int64_spectrum, esi["int64_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.uint64_spectrum, esi["uint64_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.float32_spectrum, esi["float32_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.float64_spectrum, esi["float64_spectrum"].initial
    )
    await assert_val_reading(
        everything_device.my_state_spectrum, esi["my_state_spectrum"].initial
    )

    await assert_val_reading(everything_device.bool_image, esi["bool_image"].initial)
    await assert_val_reading(everything_device.int8_image, esi["int8_image"].initial)
    await assert_val_reading(everything_device.uint8_image, esi["uint8_image"].initial)
    await assert_val_reading(everything_device.int16_image, esi["int16_image"].initial)
    await assert_val_reading(
        everything_device.uint16_image, esi["uint16_image"].initial
    )
    await assert_val_reading(everything_device.int32_image, esi["int32_image"].initial)
    await assert_val_reading(
        everything_device.uint32_image, esi["uint32_image"].initial
    )
    await assert_val_reading(everything_device.int64_image, esi["int64_image"].initial)
    await assert_val_reading(
        everything_device.uint64_image, esi["uint64_image"].initial
    )
    await assert_val_reading(
        everything_device.float32_image, esi["float32_image"].initial
    )
    await assert_val_reading(
        everything_device.float64_image, esi["float64_image"].initial
    )


@pytest.mark.asyncio
async def test_set_callback(everything_device_trl):
    source = get_full_attr_trl(everything_device_trl, "float32")
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnectedError) as exc:
        await transport.put(1.0)
    assert "Not connected" in str(exc.value)

    await transport.connect(1)
    val = None

    def callback(reading):
        nonlocal val
        val = reading["value"]

    # Correct usage
    transport.set_callback(callback)
    current_value = await transport.get_value()
    new_value = current_value + 2
    await transport.put(new_value)
    await asyncio.sleep(0.1)
    assert val == new_value

    # Try to add second callback
    with pytest.raises(RuntimeError) as exc:
        transport.set_callback(callback)
    assert "Cannot set a callback when one is already set"

    transport.set_callback(None)

    # Try to add a callback to a non-callable proxy
    transport.allow_events(False)
    transport.set_polling(False)
    with pytest.raises(RuntimeError) as exc:
        transport.set_callback(callback)
    assert "Cannot set event" in str(exc.value)

    # Try to add a non-callable callback
    transport.allow_events(True)
    with pytest.raises(RuntimeError) as exc:
        transport.set_callback(1)
    assert "Callback must be a callable" in str(exc.value)


@pytest.mark.asyncio
@pytest.mark.parametrize("allow", [True, False])
async def test_tango_transport_allow_events(everything_device_trl, allow):
    source = get_full_attr_trl(everything_device_trl, "float32")
    transport = await make_backend(float, source, connect=False)
    transport.allow_events(allow)
    assert transport.support_events == allow


@pytest.mark.asyncio
async def test_tango_transport_set_polling(everything_device_trl):
    source = get_full_attr_trl(everything_device_trl, "float32")
    transport = await make_backend(float, source, connect=False)
    transport.set_polling(True, 0.1, 1, 0.1)
    assert transport._polling == (True, 0.1, 1, 0.1)


@pytest.mark.asyncio
@pytest.mark.timeout(12.0)
async def test_attribute_subscribe_callback(everything_device_trl):
    source = get_full_attr_trl(everything_device_trl, "int64")
    backend = await make_backend(int, source)
    attr_proxy = backend.proxies[source]
    val = None

    def callback(reading):
        nonlocal val
        val = reading["value"]

    attr_proxy.subscribe_callback(callback)
    assert attr_proxy.has_subscription()
    old_value = await attr_proxy.get()
    new_value = old_value + 1
    await attr_proxy.put(new_value)
    await asyncio.sleep(0.2)
    attr_proxy.unsubscribe_callback()
    assert val == new_value

    attr_proxy.set_polling(False)
    attr_proxy.support_events = False
    with pytest.raises(RuntimeError) as exc:
        attr_proxy.subscribe_callback(callback)
    assert "Cannot set a callback" in str(exc.value)


@pytest.mark.asyncio
async def test_attribute_unsubscribe_callback(everything_device_trl):
    source = get_full_attr_trl(everything_device_trl, "float32")
    backend = await make_backend(float, source)
    attr_proxy = backend.proxies[source]

    def callback(reading):
        pass

    attr_proxy.subscribe_callback(callback)
    assert attr_proxy.has_subscription()
    attr_proxy.unsubscribe_callback()
    assert not attr_proxy.has_subscription()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
@pytest.mark.timeout(3.0)
async def test_parse_precision(everything_device_trl):
    proxy = await DeviceProxy(everything_device_trl)
    for attr in proxy.get_attribute_list():
        config = await proxy.get_attribute_config(attr)
        precision = parse_precision(config)
        if attr in [
            "float32",
            "float32_image",
            "float32_spectrum",
            "float64",
            "float64_image",
            "float64_spectrum",
        ]:
            assert precision == 2
        else:
            assert precision is None


@pytest.mark.asyncio
async def test_infer_python_type(everything_device_trl):
    proxy = await DeviceProxy(everything_device_trl)
    bad_attr = everything_device_trl + "/this_does_not_exist"
    with pytest.raises(RuntimeError) as exc:
        await infer_python_type(trl=bad_attr, proxy=proxy)
    assert "Cannot find" in str(exc.value)



from typing import Annotated as A

from bluesky.protocols import Triggerable

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DeviceVector,
    SignalR,
    SignalRW,
    SignalX,
    StandardReadable,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import EpicsDevice, PvSuffix

from ._point_detector_channel import DemoPointDetectorChannel


class DemoPointDetector(StandardReadable, EpicsDevice, Triggerable):
    """A demo detector that produces a point values based on X and Y motors."""

    acquire_time: A[SignalRW[float], PvSuffix("AcquireTime"), Format.CONFIG_SIGNAL]
    start: A[SignalX, PvSuffix("Start.PROC")]
    acquiring: A[SignalR[bool], PvSuffix("Acquiring")]
    reset: A[SignalX, PvSuffix("Reset.PROC")]

    def __init__(self, prefix: str, num_channels: int = 3, name: str = "") -> None:
        with self.add_children_as_readables():
            self.channel = DeviceVector(
                {
                    i: DemoPointDetectorChannel(f"{prefix}{i}:")
                    for i in range(1, num_channels + 1)
                }
            )
        super().__init__(prefix=prefix, name=name)

    @AsyncStatus.wrap
    async def trigger(self):
        await self.reset.trigger()
        timeout = await self.acquire_time.get_value() + DEFAULT_TIMEOUT
        await self.start.trigger(timeout=timeout)



from unittest.mock import call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adcore, adsimdetector
from ophyd_async.testing import assert_has_calls


@pytest.fixture
async def test_adsimdetector(
    static_path_provider: StaticPathProvider,
) -> adsimdetector.SimDetector:
    async with init_devices(mock=True):
        detector = adsimdetector.SimDetector("PREFIX:", static_path_provider)
    writer = detector.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(writer.file_path_exists, True)
    return detector


def test_pvs_correct(test_adsimdetector: adsimdetector.SimDetector):
    assert (
        test_adsimdetector.driver.acquire.source == "mock+ca://PREFIX:cam1:Acquire_RBV"
    )


async def test_deadtime(
    test_adsimdetector: adsimdetector.SimDetector,
):
    trigger_modes, actual_deadtime = await test_adsimdetector.get_trigger_deadtime()
    assert trigger_modes == {DetectorTrigger.INTERNAL}
    assert actual_deadtime is None


async def test_prepare_internal(
    test_adsimdetector: adsimdetector.SimDetector,
):
    await test_adsimdetector.prepare(TriggerInfo(number_of_events=11))
    assert_has_calls(
        test_adsimdetector.driver,
        [
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(11),
        ],
    )



from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from xml.etree import ElementTree as ET

import numpy as np

from ophyd_async.core import SignalR


class NDAttributeDataType(Enum):
    INT = np.dtype(np.int32).str
    INT64 = np.dtype(np.int64).str
    DOUBLE = np.dtype(np.float64).str
    STRING = "S40"


class NDAttributePvDbrType(Enum):
    DBR_SHORT = np.dtype(np.int16).str
    DBR_ENUM = np.dtype(np.int16).str
    DBR_INT = np.dtype(np.int32).str
    DBR_LONG = np.dtype(np.int32).str
    DBR_FLOAT = np.dtype(np.float32).str
    DBR_DOUBLE = np.dtype(np.float64).str
    DBR_STRING = "S40"
    DBR_CHAR = np.dtype(np.int8).str


@dataclass
class NDAttributePv:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    signal: SignalR  # caget the pv given by signal.source and attach to each frame
    dbrtype: NDAttributePvDbrType
    description: str = ""  # A description that appears in the HDF file as an attribute


@dataclass
class NDAttributeParam:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    param: str  # The parameter string as seen in the INP link of the record
    datatype: NDAttributeDataType  # The datatype of the parameter
    addr: int = 0  # The address as seen in the INP link of the record
    description: str = ""  # A description that appears in the HDF file as an attribute


def ndattributes_to_xml(
    ndattributes: Sequence[NDAttributeParam | NDAttributePv],
) -> str:
    """Convert a set of NDAttribute params to XML."""
    root = ET.Element("Attributes")
    for ndattribute in ndattributes:
        if isinstance(ndattribute, NDAttributeParam):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="PARAM",
                source=ndattribute.param,
                addr=str(ndattribute.addr),
                datatype=ndattribute.datatype.name,
                description=ndattribute.description,
            )
        elif isinstance(ndattribute, NDAttributePv):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="EPICS_PV",
                # Strip protocol prefix if present
                source=ndattribute.signal.source.split("://")[-1],
                dbrtype=ndattribute.dbrtype.name,
                description=ndattribute.description,
            )
        else:
            raise ValueError(
                f"Invalid type for ndattributes: {type(ndattribute)}. "
                "Expected NDAttributePv or NDAttributeParam."
            )
    xml_text = ET.tostring(root, encoding="unicode")
    return xml_text



from __future__ import annotations

import time
import typing
from abc import abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Generic, get_args

import numpy as np
from bluesky.protocols import Reading
from event_model import DataKey

from ._datatypes import Table
from ._signal_backend import (
    Array1D,
    EnumT,
    Primitive,
    PrimitiveT,
    SignalBackend,
    SignalDatatype,
    SignalDatatypeT,
    TableT,
    make_datakey,
    make_metadata,
)
from ._utils import Callback, cached_get_origin, get_dtype, get_enum_cls


class SoftConverter(Generic[SignalDatatypeT]):
    # This is Any -> SignalDatatypeT because we support coercing
    # value types to SignalDatatype to allow people to do things like
    # SignalRW[Enum].set("enum value")
    @abstractmethod
    def write_value(self, value: Any) -> SignalDatatypeT: ...


@dataclass
class PrimitiveSoftConverter(SoftConverter[PrimitiveT]):
    datatype: type[PrimitiveT]

    def write_value(self, value: Any) -> PrimitiveT:
        return self.datatype(value) if value else self.datatype()


class SequenceStrSoftConverter(SoftConverter[Sequence[str]]):
    def write_value(self, value: Any) -> Sequence[str]:
        return [str(v) for v in value] if value else []


@dataclass
class SequenceEnumSoftConverter(SoftConverter[Sequence[EnumT]]):
    datatype: type[EnumT]

    def write_value(self, value: Any) -> Sequence[EnumT]:
        return [self.datatype(v) for v in value] if value else []


@dataclass
class NDArraySoftConverter(SoftConverter[Array1D]):
    datatype: np.dtype | None = None

    def write_value(self, value: Any) -> Array1D:
        return np.array(() if value is None else value, dtype=self.datatype)


@dataclass
class EnumSoftConverter(SoftConverter[EnumT]):
    datatype: type[EnumT]

    def write_value(self, value: Any) -> EnumT:
        return (
            self.datatype(value)
            if value
            else list(self.datatype.__members__.values())[0]
        )


@dataclass
class TableSoftConverter(SoftConverter[TableT]):
    datatype: type[TableT]

    def write_value(self, value: Any) -> TableT:
        if isinstance(value, dict):
            return self.datatype(**value)
        elif isinstance(value, self.datatype):
            return value
        elif value is None:
            return self.datatype()
        else:
            raise TypeError(f"Cannot convert {value} to {self.datatype}")


@lru_cache
def make_converter(datatype: type[SignalDatatype]) -> SoftConverter:
    enum_cls = get_enum_cls(datatype)
    if datatype in (Sequence[str], typing.Sequence[str]):
        return SequenceStrSoftConverter()
    elif cached_get_origin(datatype) in (Sequence, typing.Sequence) and enum_cls:
        return SequenceEnumSoftConverter(enum_cls)
    elif datatype is np.ndarray:
        return NDArraySoftConverter()
    elif cached_get_origin(datatype) == np.ndarray:
        if datatype not in get_args(SignalDatatype):
            raise TypeError(f"Expected Array1D[dtype], got {datatype}")
        return NDArraySoftConverter(get_dtype(datatype))
    elif enum_cls:
        return EnumSoftConverter(enum_cls)
    elif issubclass(datatype, Table):
        return TableSoftConverter(datatype)
    elif issubclass(datatype, Primitive):
        return PrimitiveSoftConverter(datatype)
    raise TypeError(f"Can't make converter for {datatype}")


class SoftSignalBackend(SignalBackend[SignalDatatypeT]):
    """An backend to a soft Signal, for test signals see [](#MockSignalBackend).

    :param datatype: The datatype of the signal, defaults to float if not given.
    :param initial_value:
        The initial value of the signal, defaults to the "empty", "zero" or
        "default" value of the datatype if not given.
    :param units: The units for numeric datatypes.
    :param precision:
        The number of digits after the decimal place to display for a float datatype.
    """

    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        initial_value: SignalDatatypeT | None = None,
        units: str | None = None,
        precision: int | None = None,
    ):
        # Create the right converter for the datatype
        self.converter = make_converter(datatype or float)
        # Add the extra static metadata to the dictionary
        self.metadata = make_metadata(datatype, units, precision)
        # Create and set the initial value
        self.initial_value = self.converter.write_value(initial_value)
        self.reading: Reading[SignalDatatypeT]
        self.callback: Callback[Reading[SignalDatatypeT]] | None = None
        self.set_value(self.initial_value)
        super().__init__(datatype)

    def set_value(self, value: SignalDatatypeT):
        """Set the current value, alarm and timestamp."""
        self.reading = Reading(
            value=self.converter.write_value(value),
            timestamp=time.time(),
            alarm_severity=0,
        )
        if self.callback:
            self.callback(self.reading)

    def source(self, name: str, read: bool) -> str:
        return f"soft://{name}"

    async def connect(self, timeout: float):
        pass

    async def put(self, value: SignalDatatypeT | None) -> None:
        write_value = self.initial_value if value is None else value
        self.set_value(write_value)

    async def get_datakey(self, source: str) -> DataKey:
        return make_datakey(
            self.datatype or float, self.reading["value"], source, self.metadata
        )

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        return self.reading

    async def get_value(self) -> SignalDatatypeT:
        return self.reading["value"]

    async def get_setpoint(self) -> SignalDatatypeT:
        # For a soft signal, the setpoint and readback values are the same.
        return self.reading["value"]

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        if callback and self.callback:
            raise RuntimeError("Cannot set a callback when one is already set")
        if callback:
            callback(self.reading)
        self.callback = callback



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from functools import cached_property
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock

from bluesky.protocols import Reading
from event_model import DataKey

from ._derived_signal_backend import DerivedSignalBackend
from ._signal_backend import SignalBackend, SignalDatatypeT
from ._soft_signal_backend import SoftSignalBackend
from ._utils import Callback

if TYPE_CHECKING:
    from ._device import LazyMock

MockPutCallback = (
    Callable[[SignalDatatypeT], SignalDatatypeT | None]
    | Callable[[SignalDatatypeT], Awaitable[SignalDatatypeT | None]]
)


class MockSignalBackend(SignalBackend[SignalDatatypeT]):
    """Signal backend for testing, created by ``Device.connect(mock=True)``."""

    def __init__(
        self,
        initial_backend: SignalBackend[SignalDatatypeT],
        mock: LazyMock,
    ) -> None:
        if isinstance(initial_backend, MockSignalBackend):
            raise ValueError("Cannot make a MockSignalBackend for a MockSignalBackend")

        self.initial_backend = initial_backend

        if isinstance(self.initial_backend, SoftSignalBackend | DerivedSignalBackend):
            # Backend is already a SoftSignalBackend, so use it
            self.soft_backend = self.initial_backend
        else:
            # Backend is not a SoftSignalBackend, so create one to mimic it
            self.soft_backend = SoftSignalBackend(
                datatype=self.initial_backend.datatype
            )

        # use existing Mock if provided
        self.mock = mock
        self._mock_put_callback: MockPutCallback | None = None
        super().__init__(datatype=self.initial_backend.datatype)

    def set_mock_put_callback(self, callback: MockPutCallback | None):
        if "put_mock" in self.__dict__:
            # put_mock cached property exists, so set the side effect on it
            self.put_mock.side_effect = callback
        else:
            # put_mock doesn't exist, don't create it as that would be slow
            # so just keep it internally
            self._mock_put_callback = callback

    @cached_property
    def put_mock(self) -> AsyncMock:
        """Return the mock that will track calls to `put()`."""
        put_mock = AsyncMock(
            name="put",
            spec=Callable,
            side_effect=self._mock_put_callback
            if self._mock_put_callback
            else lambda v: None,
        )
        self.mock().attach_mock(put_mock, "put")
        return put_mock

    def set_value(self, value: SignalDatatypeT):
        """Set the value of the signal."""
        self.soft_backend.set_value(value)

    def source(self, name: str, read: bool) -> str:
        return f"mock+{self.initial_backend.source(name, read)}"

    async def connect(self, timeout: float) -> None:
        raise RuntimeError("It is not possible to connect a MockSignalBackend")

    @cached_property
    def put_proceeds(self) -> asyncio.Event:
        """Return an Event that will block `put()` until set.

        The Event is initially set, but can be unset to block `put()`.
        """
        put_proceeds = asyncio.Event()
        put_proceeds.set()
        return put_proceeds

    async def put(self, value: SignalDatatypeT | None):
        new_value = await self.put_mock(value)
        if new_value is None:
            new_value = value
        await self.soft_backend.put(new_value)
        await self.put_proceeds.wait()

    async def get_reading(self) -> Reading:
        return await self.soft_backend.get_reading()

    async def get_value(self) -> SignalDatatypeT:
        return await self.soft_backend.get_value()

    async def get_setpoint(self) -> SignalDatatypeT:
        return await self.soft_backend.get_setpoint()

    async def get_datakey(self, source: str) -> DataKey:
        return await self.soft_backend.get_datakey(source)

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        self.soft_backend.set_callback(callback)



import asyncio
from contextlib import suppress

from ophyd_async.core import DetectorArmLogic, TriggerInfo

from ._pattern_generator import PatternGenerator


class BlobArmLogic(DetectorArmLogic):
    def __init__(self, pattern_generator: PatternGenerator):
        self.pattern_generator = pattern_generator
        self.trigger_info: TriggerInfo | None = None
        self.task: asyncio.Task | None = None

    async def arm(self):
        # Start a background process off writing the images to file
        self.task = asyncio.create_task(self.pattern_generator.write_images_to_file())

    async def wait_for_idle(self):
        # Wait for the background task to complete
        if self.task:
            await self.task

    async def disarm(self):
        # Stop the background task and wait for it to finish
        if self.task:
            self.task.cancel()
            with suppress(asyncio.CancelledError):
                await self.task
            self.task = None



import asyncio

from ophyd_async.core import DetectorTriggerLogic, EnableDisable

from ._io import ADBaseIO, ADImageMode, NDCBFlushOnSoftTrgMode, NDCircularBuffIO


async def prepare_exposures(
    driver: ADBaseIO,
    num: int,
    livetime: float = 0.0,
    deadtime: float = 0.0,
):
    image_mode = ADImageMode.CONTINUOUS if num == 0 else ADImageMode.MULTIPLE
    coros = [
        driver.image_mode.set(image_mode),
        driver.num_images.set(num),
    ]
    if livetime:
        coros.append(driver.acquire_time.set(livetime))
        if deadtime:
            coros.append(driver.acquire_period.set(livetime + deadtime))
    await asyncio.gather(*coros)


class ADContAcqTriggerLogic(DetectorTriggerLogic):
    def __init__(self, driver: ADBaseIO, cb_plugin: NDCircularBuffIO):
        self.driver = driver
        self.cb_plugin = cb_plugin

    async def _ensure_driver_acquiring(self, livetime: float):
        # Check the current state of the system
        image_mode, acquiring, acquire_time = await asyncio.gather(
            self.driver.image_mode.get_value(),
            self.driver.acquire.get_value(),
            self.driver.acquire_time.get_value(),
        )
        # Ensure the detector is in continuous acquisition mode
        if image_mode != ADImageMode.CONTINUOUS or not acquiring:
            raise RuntimeError(
                "Driver must be acquiring in continuous mode to use the "
                "cont acq interface"
            )
        # Not all detectors allow for changing exposure times during an acquisition,
        # so in this least-common-denominator implementation check to see if
        # exposure time matches the current exposure time.
        if livetime and livetime != acquire_time:
            raise ValueError(
                f"Detector exposure time currently set to {acquire_time}, "
                f"but requested exposure is {livetime}"
            )

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await self._ensure_driver_acquiring(livetime)
        # Setup the CB plugin with the current parameters
        await asyncio.gather(
            self.cb_plugin.enable_callbacks.set(EnableDisable.ENABLE),
            self.cb_plugin.pre_count.set(0),
            self.cb_plugin.post_count.set(num),
            self.cb_plugin.preset_trigger_count.set(1),
            self.cb_plugin.flush_on_soft_trg.set(NDCBFlushOnSoftTrgMode.ON_NEW_IMAGE),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ophyd_async.core import Device, DeviceConnector, DeviceFiller

from ._signal import get_signal_backend_type, split_protocol_from_pv
from ._util import EpicsOptions, EpicsSignalBackend


@dataclass
class PvSuffix:
    """Define the PV suffix to be appended to the device prefix.

    For a SignalRW:
    - If you use the same "Suffix" for the read and write PV then use PvSuffix("Suffix")
    - If you have "Suffix" for the write PV and "Suffix_RBV" for the read PV then use
      PvSuffix.rbv("Suffix")
    - If you have "WriteSuffix" for the write PV and "ReadSuffix" for the read PV then
      you use PvSuffix(read_suffix="ReadSuffix", write_suffix="WriteSuffix")

    For a SignalR:
    - If you have "Suffix" for the read PV then use PvSuffix("Suffix")
    - If you have "Suffix_RBV" for the read PV then use PvSuffix("Suffix_RBV"), do not
      use PvSuffix.rbv as that will try to connect to multiple PVs
    """

    read_suffix: str
    write_suffix: str | None = None

    @classmethod
    def rbv(cls, write_suffix: str, rbv_suffix: str = "_RBV") -> PvSuffix:
        return cls(write_suffix + rbv_suffix, write_suffix)


def fill_backend_with_prefix(
    prefix: str, backend: EpicsSignalBackend, annotations: list[Any]
):
    unhandled = []
    while annotations:
        annotation = annotations.pop(0)
        if isinstance(annotation, PvSuffix):
            backend.read_pv = prefix + annotation.read_suffix
            backend.write_pv = prefix + (
                annotation.write_suffix or annotation.read_suffix
            )
        elif isinstance(annotation, EpicsOptions):
            backend.options = annotation
        else:
            unhandled.append(annotation)
    annotations.extend(unhandled)
    # These leftover annotations will now be handled by the iterator


class EpicsDeviceConnector(DeviceConnector):
    """Used for connecting signals to static EPICS pvs."""

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            protocol, prefix = split_protocol_from_pv(self.prefix)
            self.filler = DeviceFiller(
                device,
                signal_backend_factory=get_signal_backend_type(protocol),
                device_connector_factory=DeviceConnector,
            )
            for backend, annotations in self.filler.create_signals_from_annotations():
                fill_backend_with_prefix(prefix, backend, annotations)

            list(self.filler.create_devices_from_annotations())



from functools import reduce

import numpy as np
import pytest
from pydantic import ValidationError

from ophyd_async.fastcs.panda import SeqTable, SeqTrigger


def test_seq_table_converts_lists():
    seq_table_dict_with_lists = {field_name: [] for field_name, _ in SeqTable()}
    # Validation passes
    seq_table = SeqTable(**seq_table_dict_with_lists)
    for field_name, field_value in seq_table:
        if field_name == "trigger":
            assert field_value == []
        else:
            assert np.array_equal(field_value, np.array([], dtype=field_value.dtype))


def test_seq_table_validation_errors():
    with pytest.raises(ValidationError, match="17 validation errors for SeqTable"):
        SeqTable(
            repeats=0,
            trigger=SeqTrigger.IMMEDIATE,
            position=0,
            time1=0,
            outa1=False,
            outb1=False,
            outc1=False,
            outd1=False,
            oute1=False,
            outf1=False,
            time2=0,
            outa2=False,
            outb2=False,
            outc2=False,
            outd2=False,
            oute2=False,
            outf2=False,
        )

    large_seq_table = SeqTable(
        repeats=np.zeros(4096, dtype=np.uint16),
        trigger=[SeqTrigger.IMMEDIATE] * 4096,
        position=np.zeros(4096, dtype=np.int32),
        time1=np.zeros(4096, dtype=np.uint32),
        outa1=np.zeros(4096, dtype=np.bool_),
        outb1=np.zeros(4096, dtype=np.bool_),
        outc1=np.zeros(4096, dtype=np.bool_),
        outd1=np.zeros(4096, dtype=np.bool_),
        oute1=np.zeros(4096, dtype=np.bool_),
        outf1=np.zeros(4096, dtype=np.bool_),
        time2=np.zeros(4096, dtype=np.uint32),
        outa2=np.zeros(4096, dtype=np.bool_),
        outb2=np.zeros(4096, dtype=np.bool_),
        outc2=np.zeros(4096, dtype=np.bool_),
        outd2=np.zeros(4096, dtype=np.bool_),
        oute2=np.zeros(4096, dtype=np.bool_),
        outf2=np.zeros(4096, dtype=np.bool_),
    )
    with pytest.raises(
        ValidationError,
        match=(
            "1 validation error for SeqTable\n  Value error, Length 4097 is too long."
        ),
    ):
        large_seq_table + SeqTable.row()
    # TODO: validation of numpy types disabled until bool -> uint8 CA issue resolved
    # with pytest.raises(
    #     ValidationError,
    #     match="1 validation error for SeqTable\n  Assertion failed, repeats: "
    #     + "expected dtype uint16, got int32",
    # ):
    #     row_one = SeqTable.row()
    #     wrong_types = {
    #         field_name: field_value.astype(np.int32)
    #         if isinstance(field_value, np.ndarray)
    #         else field_value
    #         for field_name, field_value in row_one
    #     }
    #     SeqTable(**wrong_types)
    with pytest.raises(
        ValidationError,
        match="trigger.0\n  Input should be 'Immediate', 'BITA=0'",
    ):
        SeqTable.row(trigger="A")


def test_seq_table_pva_conversion():
    pva_dict = {
        "repeats": np.array([1, 2, 3, 4], dtype=np.uint16),
        "trigger": [
            SeqTrigger.IMMEDIATE,
            SeqTrigger.IMMEDIATE,
            SeqTrigger.BITC_0,
            SeqTrigger.IMMEDIATE,
        ],
        "position": np.array([1, 2, 3, 4], dtype=np.int32),
        "time1": np.array([1, 0, 1, 0], dtype=np.uint32),
        "outa1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outb1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outc1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outd1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "oute1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outf1": np.array([1, 0, 1, 0], dtype=np.bool_),
        "time2": np.array([1, 2, 3, 4], dtype=np.uint32),
        "outa2": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outb2": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outc2": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outd2": np.array([1, 0, 1, 0], dtype=np.bool_),
        "oute2": np.array([1, 0, 1, 0], dtype=np.bool_),
        "outf2": np.array([1, 0, 1, 0], dtype=np.bool_),
    }
    row_wise_dicts = [
        {
            "repeats": 1,
            "trigger": SeqTrigger.IMMEDIATE,
            "position": 1,
            "time1": 1,
            "outa1": 1,
            "outb1": 1,
            "outc1": 1,
            "outd1": 1,
            "oute1": 1,
            "outf1": 1,
            "time2": 1,
            "outa2": 1,
            "outb2": 1,
            "outc2": 1,
            "outd2": 1,
            "oute2": 1,
            "outf2": 1,
        },
        {
            "repeats": 2,
            "trigger": SeqTrigger.IMMEDIATE,
            "position": 2,
            "time1": 0,
            "outa1": 0,
            "outb1": 0,
            "outc1": 0,
            "outd1": 0,
            "oute1": 0,
            "outf1": 0,
            "time2": 2,
            "outa2": 0,
            "outb2": 0,
            "outc2": 0,
            "outd2": 0,
            "oute2": 0,
            "outf2": 0,
        },
        {
            "repeats": 3,
            "trigger": SeqTrigger.BITC_0,
            "position": 3,
            "time1": 1,
            "outa1": 1,
            "outb1": 1,
            "outc1": 1,
            "outd1": 1,
            "oute1": 1,
            "outf1": 1,
            "time2": 3,
            "outa2": 1,
            "outb2": 1,
            "outc2": 1,
            "outd2": 1,
            "oute2": 1,
            "outf2": 1,
        },
        {
            "repeats": 4,
            "trigger": SeqTrigger.IMMEDIATE,
            "position": 4,
            "time1": 0,
            "outa1": 0,
            "outb1": 0,
            "outc1": 0,
            "outd1": 0,
            "oute1": 0,
            "outf1": 0,
            "time2": 4,
            "outa2": 0,
            "outb2": 0,
            "outc2": 0,
            "outd2": 0,
            "oute2": 0,
            "outf2": 0,
        },
    ]

    def _assert_col_equal(column1, column2):
        if isinstance(column1, np.ndarray):
            assert np.array_equal(column1, column2)
            assert column1.dtype == column2.dtype
        else:
            assert column1 == column2
            assert all(isinstance(x, SeqTrigger) for x in column1)
            assert all(isinstance(x, SeqTrigger) for x in column2)

    seq_table_from_pva_dict = SeqTable(**pva_dict)
    for (_, column1), column2 in zip(
        seq_table_from_pva_dict, pva_dict.values(), strict=False
    ):
        _assert_col_equal(column1, column2)

    seq_table_from_rows = reduce(
        lambda x, y: x + y,
        [SeqTable.row(**row_kwargs) for row_kwargs in row_wise_dicts],
    )
    for (_, column1), column2 in zip(
        seq_table_from_rows, pva_dict.values(), strict=False
    ):
        _assert_col_equal(column1, column2)

    # Idempotency
    applied_twice_to_pva_dict = SeqTable(**pva_dict).model_dump(mode="python")
    for column1, column2 in zip(
        applied_twice_to_pva_dict.values(), pva_dict.values(), strict=False
    ):
        _assert_col_equal(column1, column2)

    dtype = seq_table_from_pva_dict.numpy_dtype()
    assert dtype == np.dtype(
        [
            ("repeats", np.uint16),
            ("trigger", "|S40"),
            ("position", np.int32),
            ("time1", np.uint32),
            ("outa1", np.bool_),
            ("outb1", np.bool_),
            ("outc1", np.bool_),
            ("outd1", np.bool_),
            ("oute1", np.bool_),
            ("outf1", np.bool_),
            ("time2", np.uint32),
            ("outa2", np.bool_),
            ("outb2", np.bool_),
            ("outc2", np.bool_),
            ("outd2", np.bool_),
            ("oute2", np.bool_),
            ("outf2", np.bool_),
        ]
    )

    assert np.array_equal(
        seq_table_from_pva_dict.numpy_table(),
        np.array(
            [
                (
                    1,
                    "Immediate",
                    1,
                    1,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    1,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ),
                (
                    2,
                    "Immediate",
                    2,
                    0,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    2,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ),
                (
                    3,
                    "BITC=0",
                    3,
                    1,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    3,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ),
                (
                    4,
                    "Immediate",
                    4,
                    0,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    4,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ),
            ],
            dtype=dtype,
        ),
    )


def test_seq_table_takes_trigger_enum_row():
    table = SeqTable.row(trigger=SeqTrigger.BITA_0)
    assert table.trigger[0] == SeqTrigger.BITA_0
    table = SeqTable(
        repeats=np.array([1], dtype=np.uint16),
        trigger=[SeqTrigger.BITA_0],
        position=np.array([1], dtype=np.int32),
        time1=np.array([1], dtype=np.uint32),
        outa1=np.array([1], dtype=np.bool_),
        outb1=np.array([1], dtype=np.bool_),
        outc1=np.array([1], dtype=np.bool_),
        outd1=np.array([1], dtype=np.bool_),
        oute1=np.array([1], dtype=np.bool_),
        outf1=np.array([1], dtype=np.bool_),
        time2=np.array([1], dtype=np.uint32),
        outa2=np.array([1], dtype=np.bool_),
        outb2=np.array([1], dtype=np.bool_),
        outc2=np.array([1], dtype=np.bool_),
        outd2=np.array([1], dtype=np.bool_),
        oute2=np.array([1], dtype=np.bool_),
        outf2=np.array([1], dtype=np.bool_),
    )
    assert table.trigger[0] == SeqTrigger.BITA_0



from unittest.mock import AsyncMock, call, patch

import numpy as np
import pytest
from scanspec.specs import Fly, Line

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    get_mock,
    set_and_wait_for_value,
    set_mock_value,
)
from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac import PmacIO
from ophyd_async.epics.pmac._pmac_trajectory import (  # noqa: PLC2701
    PmacExecuteState,
    PmacTrajectoryTriggerLogic,
)
from ophyd_async.epics.pmac._utils import (  # noqa: PLC2701
    _PmacMotorInfo,
)


async def test_pmac_prepare(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, sim_x_motor, _ = sim_motors
    spec = Fly(2.0 @ Line(sim_x_motor, 1, 5, 2))
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    await pmac_trajectory.prepare(spec)

    assert await pmac_io.coord[1].cs_axis_setpoint[7].get_value() == -1.2

    assert await pmac_io.trajectory.positions[7].get_value() == pytest.approx(
        [-1.0, 1.0, 3.0, 5.0, 7.0, 7.2]
    )

    assert await pmac_io.trajectory.velocities[7].get_value() == pytest.approx(
        [2.0, 2.0, 2.0, 2.0, 2.0, 0]
    )

    assert await pmac_io.trajectory.time_array.get_value() == pytest.approx(
        [200000, 1000000, 1000000, 1000000, 1000000, 200000]
    )

    assert await pmac_io.trajectory.points_to_build.get_value() == 6


async def test_pmac_move_to_start(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, sim_x_motor, sim_y_motor = sim_motors
    motor_info = _PmacMotorInfo(
        "CS1",
        1,
        {sim_x_motor: 7, sim_y_motor: 8},
        {sim_x_motor: 10, sim_y_motor: 20},
        {sim_x_motor: 10, sim_y_motor: 10},
    )
    coord = pmac_io.coord[motor_info.cs_number]
    ramp_up_position = {sim_x_motor: np.float64(-1.2), sim_y_motor: np.float64(-0.6)}
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)

    # Wrap set_and_wait_for_value to check passed arguments
    with patch(
        "ophyd_async.epics.pmac._pmac_trajectory.set_and_wait_for_value",
        wraps=set_and_wait_for_value,
    ) as spy_set_and_wait_for_value:
        await pmac_trajectory._move_to_start(motor_info, ramp_up_position)

        coord_mock_calls = get_mock(coord).mock_calls

        assert coord_mock_calls[0] == call.defer_moves.put(True)
        assert coord_mock_calls[1] == (
            "cs_axis_setpoint.7.put",
            (np.float64(-1.2)),
            {},
        )
        assert coord_mock_calls[2] == (
            "cs_axis_setpoint.8.put",
            (np.float64(-0.6)),
            {},
        )
        assert coord_mock_calls[3] == call.defer_moves.put(False)

        # Longest move time should be sim_motor_x, so calculate this
        expected_timeout = DEFAULT_TIMEOUT + (
            abs(ramp_up_position[sim_x_motor])
            / motor_info.motor_max_velocity[sim_x_motor]
        )

        # All motors should have the same move timeout
        assert all(
            call_.kwargs["set_timeout"] == expected_timeout
            for call_ in spy_set_and_wait_for_value.mock_calls
        )


async def test_pmac_trajectory_kickoff(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    pmac_io, sim_x_motor, sim_y_motor = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    spec = Fly(2.0 @ (Line(sim_y_motor, 1, 5, 2) * ~Line(sim_x_motor, 1, 5, 2)))
    with patch("ophyd_async.epics.pmac._pmac_trajectory.SLICE_SIZE", 2):
        # This will prepare the buffer with 2 frames of info
        await pmac_trajectory.prepare(spec)
        # This will consume another 2 frames
        set_mock_value(
            pmac_io.trajectory.total_points, 2
        )  # Only one value in observe_value(total_points)
        await pmac_trajectory.kickoff()  # Executes trajectory, appending once
        await pmac_trajectory.complete()  # Block until trajectory is complete

    sim_y_motor_position_arrays = [
        np.array(call.args[0])
        for call in get_mock(pmac_io.trajectory.positions[7]).mock_calls
    ]

    sim_x_motor_position_arrays = [
        np.array(call.args[0])
        for call in get_mock(pmac_io.trajectory.positions[8]).mock_calls
    ]

    # Appended to buffer once in prepare and once after kickoff
    assert len(sim_x_motor_position_arrays) == len(sim_y_motor_position_arrays) == 2

    assert np.concatenate(sim_y_motor_position_arrays) == pytest.approx(
        [
            -1.0,
            1.0,
            3.0,
            5.0,
            7.0,
            7.2,
            7.2,
            7.2,
            7.0,
            5.0,
            3.0,
            1.0,
            -1,
            -1.2,
        ]
    )
    assert np.concatenate(sim_x_motor_position_arrays) == pytest.approx(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.395,
            3.0,
            4.605,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ],
        1e-3,
    )

    sim_y_motor_velocity_arrays = [
        np.array(call.args[0])
        for call in get_mock(pmac_io.trajectory.velocities[7]).mock_calls
    ]

    sim_x_motor_velocity_arrays = [
        np.array(call.args[0])
        for call in get_mock(pmac_io.trajectory.velocities[8]).mock_calls
    ]

    assert np.concatenate(sim_y_motor_velocity_arrays) == pytest.approx(
        [
            2.0,
            2.0,
            2.0,
            2.0,
            2.0,
            0.0,
            0.0,
            0.0,
            -2.0,
            -2.0,
            -2.0,
            -2.0,
            -2.0,
            0.0,
        ]
    )
    assert np.concatenate(sim_x_motor_velocity_arrays) == pytest.approx(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            3.951,
            8.888,
            3.951,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        1e-3,
    )


async def test_pmac_trajectory_kickoff_trajectory_raises_exception_if_no_prepare(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    pmac_io, _, _ = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    with pytest.raises(RuntimeError, match="Cannot kickoff. Must call prepare first."):
        await pmac_trajectory.kickoff()


async def test_pmac_trajectory_complete(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, _, _ = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    with pytest.raises(RuntimeError, match="Cannot complete. Must call kickoff first."):
        await pmac_trajectory.complete()


async def test_pmac_trajectory_stage(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, _, _ = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    mock_pmac_trajectory_io = get_mock(pmac_trajectory.pmac.trajectory)
    await pmac_trajectory.stage()

    # Check that all axes are then set not be used
    assert all(
        get_mock(axis).put.assert_called_once_with(False) is None
        for axis in pmac_trajectory.pmac.trajectory.use_axis.values()
    )

    # Check that an empty trajectory is then executed
    assert mock_pmac_trajectory_io.mock_calls[
        len(pmac_trajectory.pmac.trajectory.use_axis) :
    ] == [
        call.time_array.put(np.array(0)),
        call.user_array.put(np.array(8)),
        call.points_to_build.put(1),
        call.build_profile.put(None),
        call.execute_profile.put(True),
    ]


async def test_pmac_trajectory_unstage(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, _, _ = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    pmac_trajectory._stop_if_running = AsyncMock()
    await pmac_trajectory.unstage()
    pmac_trajectory._stop_if_running.assert_called_once()


async def test_trajectory_stop_if_running(sim_motors: tuple[PmacIO, Motor, Motor]):
    pmac_io, _, _ = sim_motors
    pmac_trajectory = PmacTrajectoryTriggerLogic(pmac_io)
    mock_abort_profile = get_mock(pmac_io.trajectory.abort_profile)
    mock_abort_profile.put = AsyncMock()

    # Method not called as no running trajectory
    await pmac_trajectory._stop_if_running()
    mock_abort_profile.put.assert_not_called()

    # Mocking that trajectory is executing
    set_mock_value(
        pmac_trajectory.pmac.trajectory.execute_state, PmacExecuteState.EXECUTING
    )

    # Method called as there is now a running trajectory
    await pmac_trajectory._stop_if_running()
    mock_abort_profile.put.assert_called_once()



from collections.abc import Sequence

import bluesky.plan_stubs as bps
from bluesky.utils import plan

from ._detector import AreaDetector
from ._io import NDArrayBaseIO, NDStatsIO
from ._ndattribute import (
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    ndattributes_to_xml,
)


@plan
def setup_ndattributes(
    device: NDArrayBaseIO,
    ndattributes: Sequence[NDAttributeParam | NDAttributePv],
):
    xml = ndattributes_to_xml(ndattributes)
    yield from bps.abs_set(
        device.nd_attributes_file,
        xml,
        wait=True,
    )


@plan
def setup_ndstats_sum(detector: AreaDetector, stats_name: str = "stats"):
    """Set up nd stats sum nd attribute for a detector."""
    stats = detector.get_plugin(stats_name, NDStatsIO)
    yield from (
        setup_ndattributes(
            stats,
            [
                NDAttributeParam(
                    name=f"{detector.name}-sum",
                    param="TOTAL",
                    datatype=NDAttributeDataType.DOUBLE,
                    description="Sum of the array",
                )
            ],
        )
    )



from ._epics_connector import EpicsDeviceConnector, PvSuffix
from ._epics_device import EpicsDevice
from ._pvi_connector import PviDeviceConnector, PviTree, SignalDetails
from ._signal import (
    CaSignalBackend,
    PvaSignalBackend,
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    epics_signal_w,
    epics_signal_x,
)
from ._util import EpicsOptions, stop_busy_record, wait_for_good_state

__all__ = [
    "PviDeviceConnector",
    "PviTree",
    "SignalDetails",
    "EpicsDeviceConnector",
    "PvSuffix",
    "EpicsDevice",
    "CaSignalBackend",
    "PvaSignalBackend",
    "epics_signal_r",
    "epics_signal_rw",
    "epics_signal_rw_rbv",
    "epics_signal_w",
    "epics_signal_x",
    "stop_busy_record",
    "wait_for_good_state",
    "EpicsOptions",
]



from unittest.mock import call

import pytest

from ophyd_async.core import get_mock, init_devices
from ophyd_async.epics import adcore


@pytest.fixture
async def adbase():
    async with init_devices(mock=True):
        adbase = adcore.ADBaseIO("PREFIX:ADBASE:")
    return adbase


@pytest.mark.parametrize(
    "num,livetime,deadtime,expected_calls",
    [
        # Single exposure with default livetime and deadtime
        (
            1,
            0.0,
            0.0,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(1),
            ],
        ),
        # Multiple exposures with no livetime or deadtime
        (
            5,
            0.0,
            0.0,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(5),
            ],
        ),
        # Continuous mode (num=0)
        (
            0,
            0.0,
            0.0,
            [
                call.image_mode.put(adcore.ADImageMode.CONTINUOUS),
                call.num_images.put(0),
            ],
        ),
        # With livetime only
        (
            5,
            0.1,
            0.0,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(5),
                call.acquire_time.put(0.1),
            ],
        ),
        # With livetime and deadtime
        (
            10,
            0.2,
            0.05,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(10),
                call.acquire_time.put(0.2),
                call.acquire_period.put(0.25),
            ],
        ),
        # Large number of exposures with livetime only
        (
            100,
            0.01,
            0.0,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(100),
                call.acquire_time.put(0.01),
            ],
        ),
        # With deadtime but no livetime (deadtime should be ignored)
        (
            5,
            0.0,
            0.1,
            [
                call.image_mode.put(adcore.ADImageMode.MULTIPLE),
                call.num_images.put(5),
            ],
        ),
    ],
)
async def test_prepare_exposures(
    adbase: adcore.ADBaseIO, num, livetime, deadtime, expected_calls
):
    await adcore.prepare_exposures(
        adbase, num=num, livetime=livetime, deadtime=deadtime
    )
    assert get_mock(adbase).mock_calls == expected_calls



from ophyd_async.core import DEFAULT_TIMEOUT, DetectorArmLogic, wait_for_value

from ._io import DetectorStatus, JungfrauDriverIO, PedestalMode


class JungfrauArmLogic(DetectorArmLogic):
    def __init__(self, detector: JungfrauDriverIO) -> None:
        self.detector = detector

    async def arm(self):
        await self.detector.acquisition_start.trigger()

    async def wait_for_idle(self):
        await wait_for_value(
            self.detector.detector_status, DetectorStatus.IDLE, timeout=DEFAULT_TIMEOUT
        )

    async def disarm(self):
        await self.detector.acquisition_stop.trigger()
        await self.detector.pedestal_mode_state.set(PedestalMode.OFF)
